div.sphinxsidebar p.caption {
    font-weight: 300;
    font-size: 1.4rem;
}

details {
    margin-block-start: 1em;
    margin-block-end: 1em;
}

div.admonition {
    padding-bottom: 0;
}

div.admonition p.admonition-title {
    font-size: 17px;
    font-weight: bold;
}

div.tip {
    background-color: #DFD;
    border-color: #ACA;
}

div.scroll-y pre {
    max-height: 20em;
    overflow-y: auto;
}

div.tutorial pre {
    border-left: 5px solid #5A5;
}

@media screen and (min-width: 875px) {
    div.document {
        width: 100%;
    }
}
@media screen and (min-width: 1095px) {
    div.document {
        width: 1095px;
    }
}
